#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import logging
import typing
from contextlib import contextmanager
from typing import Any, Iterator, Mapping, Optional, Sequence

from opentelemetry import trace as trace_api
from opentelemetry.sdk import trace as oteltrace
from opentelemetry.trace import Context, Link, SpanKind
from opentelemetry.trace.propagation import _SPAN_KEY
from opentelemetry.trace.status import Status, StatusCode
from opentelemetry.util import types

import elasticapm
from elasticapm import Client
from elasticapm.traces import execution_context

from . import context as context_api
from .span import Span
from .utils import get_span_kind, get_traceparent

logger = logging.getLogger("elasticapm.otel")


class Tracer(oteltrace.Tracer):
    """
    Handles span creation and in-process context propagation.
    This class provides methods for manipulating the context, creating spans,
    and controlling spans' lifecycles.
    """

    def __init__(
        self, *args, elasticapm_client: Optional[Client] = None, config: Optional[Mapping] = None, **kwargs
    ) -> None:  # type: ignore
        self.client = elasticapm_client
        if not self.client:
            self.client = elasticapm.get_client()
        if not self.client:
            self.client = elasticapm.Client(config=config)
        if self.client.config.instrument and self.client.config.enabled:
            elasticapm.instrument()

    def start_span(
        self,
        name: str,
        context: Optional[Context] = None,
        kind: SpanKind = SpanKind.INTERNAL,
        attributes: types.Attributes = None,
        links: Optional[Sequence[Link]] = None,
        start_time: Optional[int] = None,
        record_exception: bool = True,
        set_status_on_exception: bool = True,
    ) -> "Span":
        """
        Starts a span.
        Create a new span. Start the span without setting it as the current
        span in the context. To start the span and use the context in a single
        method, see :meth:`start_as_current_span`.
        By default the current span in the context will be used as parent, but an
        explicit context can also be specified, by passing in a `Context` containing
        a current `Span`. If there is no current span in the global `Context` or in
        the specified context, the created span will be a root span.
        The span can be used as a context manager. On exiting the context manager,
        the span's end() method will be called.
        Example::
            # trace.get_current_span() will be used as the implicit parent.
            # If none is found, the created span will be a root instance.
            with tracer.start_span("one") as child:
                child.add_event("child's event")
        Args:
            name: The name of the span to be created.
            context: An optional Context containing the span's parent. Defaults to the
                global context.
            kind: The span's kind (relationship to parent). Note that is
                meaningful even if there is no parent.
            attributes: The span's attributes.
            links: Links span to other spans (ignored in this bridge)
            start_time: Sets the start time of a span
            record_exception: Whether to record any exceptions raised within the
                context as error event on the span. (ignored in this bridge)
            set_status_on_exception: Only relevant if the returned span is used
                in a with/context manager. Defines wether the span status will
                be automatically set to ERROR when an uncaught exception is
                raised in the span with block. The span status won't be set by
                this mechanism if it was previously set manually.
        Returns:
            The newly-created span.
        """
        if not record_exception:
            logger.warning("record_exception was set to False, but exceptions will still be recorded for this span.")

        parent_span_context = trace_api.get_current_span(context).get_span_context()
        if parent_span_context is not None and not isinstance(parent_span_context, trace_api.SpanContext):
            raise TypeError("parent_span_context must be a SpanContext or None.")
        traceparent = get_traceparent(parent_span_context)

        span = None
        current_transaction = execution_context.get_transaction()
        client = self.client

        elastic_links = tuple(get_traceparent(link.context) for link in links) if links else None
        if traceparent and current_transaction:
            logger.warning(
                "Remote context included when a transaction was already active. "
                "Ignoring remote context and creating a Span instead."
            )
        elif traceparent:
            elastic_span = client.begin_transaction(
                "otel", trace_parent=traceparent, start=start_time, auto_activate=False, links=elastic_links
            )
            span = Span(
                name=name,
                elastic_span=elastic_span,
                set_status_on_exception=set_status_on_exception,
                client=self.client,
            )
            span.set_attributes(attributes)
        elif not current_transaction:
            elastic_span = client.begin_transaction("otel", start=start_time, auto_activate=False, links=elastic_links)
            span = Span(
                name=name,
                elastic_span=elastic_span,
                set_status_on_exception=set_status_on_exception,
                client=self.client,
            )
            span.set_attributes(attributes)
        else:
            elastic_span = current_transaction.begin_span(
                name, "otel", start=start_time, auto_activate=False, links=elastic_links
            )
            span = Span(
                name=name,
                elastic_span=elastic_span,
                set_status_on_exception=set_status_on_exception,
                client=self.client,
            )
            span.set_attributes(attributes)
        spankind = get_span_kind(kind)
        elastic_span.context["otel_spankind"] = spankind

        return span

    @contextmanager  # type: ignore
    def start_as_current_span(
        self,
        name: str,
        context: Optional[Context] = None,
        kind: SpanKind = SpanKind.INTERNAL,
        attributes: types.Attributes = None,
        links: Optional[Sequence[Link]] = None,
        start_time: Optional[int] = None,
        record_exception: bool = True,
        set_status_on_exception: bool = True,
        end_on_exit: bool = True,
    ) -> Iterator["Span"]:
        """
        Context manager for creating a new span and set it
        as the current span in this tracer's context.
        Exiting the context manager will call the span's end method,
        as well as return the current span to its previous value by
        returning to the previous context.
        Example::
            with tracer.start_as_current_span("one") as parent:
                parent.add_event("parent's event")
                with trace.start_as_current_span("two") as child:
                    child.add_event("child's event")
                    trace.get_current_span()  # returns child
                trace.get_current_span()      # returns parent
            trace.get_current_span()          # returns previously active span
        This is a convenience method for creating spans attached to the
        tracer's context. Applications that need more control over the span
        lifetime should use :meth:`start_span` instead. For example::
            with tracer.start_as_current_span(name) as span:
                do_work()
        is equivalent to::
            span = tracer.start_span(name)
            with opentelemetry.trace.use_span(span, end_on_exit=True):
                do_work()
        Args:
            name: The name of the span to be created.
            context: An optional Context containing the span's parent. Defaults to the
                global context.
            kind: The span's kind (relationship to parent). Note that is
                meaningful even if there is no parent.
            attributes: The span's attributes.
            links: Links span to other spans (ignored in this bridge)
            start_time: Sets the start time of a span
            record_exception: Whether to record any exceptions raised within the
                context as error event on the span. (ignored in this bridge)
            set_status_on_exception: Only relevant if the returned span is used
                in a with/context manager. Defines wether the span status will
                be automatically set to ERROR when an uncaught exception is
                raised in the span with block. The span status won't be set by
                this mechanism if it was previously set manually.
            end_on_exit: Whether to end the span automatically when leaving the
                context manager.
        Yields:
            The newly-created span.
        """
        span = self.start_span(
            name=name,
            context=context,
            kind=kind,
            attributes=attributes,
            links=links,
            start_time=start_time,
            record_exception=record_exception,
            set_status_on_exception=set_status_on_exception,
        )
        with use_span(
            span,
            end_on_exit=end_on_exit,
            record_exception=record_exception,
            set_status_on_exception=set_status_on_exception,
        ) as activated_span:
            yield activated_span


def get_tracer(
    instrumenting_module_name: str,
    instrumenting_library_version: typing.Optional[str] = None,
    tracer_provider: Optional[Any] = None,
    schema_url: Optional[str] = None,
    elasticapm_client: Optional[Client] = None,
    config: Optional[Mapping] = None,
) -> "Tracer":
    """
    Returns the Elastic-wrapped Tracer object which allows for span creation

    Args:
        instrumenting_module_name:  The name of the instrumenting module
            (usually just `__name__`)

    All other args are ignored in this implementation.
    """
    return Tracer(instrumenting_module_name, elasticapm_client=elasticapm_client, config=config)


def set_tracer_provider(tracer_provider: Any) -> None:
    """
    No-op to match opentelemetry's `trace` module
    """
    return None


def get_tracer_provider() -> None:
    """
    Not implemented by otel bridge
    """
    raise NotImplementedError()


@contextmanager
def use_span(
    span: Span,
    end_on_exit: bool = False,
    record_exception: bool = True,
    set_status_on_exception: bool = None,
) -> None:
    """
    Takes a non-active span and activates it in the current context.
    """
    if set_status_on_exception is None:
        set_status_on_exception = span.set_status_on_exception
    if set_status_on_exception is None:
        # Default if it's not set in the span or in this context manager
        set_status_on_exception = True
    context_api.attach(context_api.set_value(_SPAN_KEY, span))
    try:
        yield span
    except Exception as exc:
        if record_exception:
            elasticapm.get_client().capture_exception(handled=False)
        if set_status_on_exception:
            span.set_status(
                Status(
                    status_code=StatusCode.ERROR,
                    description=f"{type(exc).__name__}: {exc}",
                )
            )
        raise
    finally:
        if end_on_exit:
            span.end()
        else:
            # Spans auto-detach when they end.
            context_api.detach()



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import gzip
import io
import itertools
import json
import logging
import logging.handlers
import os
import random
import socket
import socketserver
import sys
import tempfile
import time
import warnings
import zlib
from collections import defaultdict
from typing import Optional
from urllib.request import pathname2url

import jsonschema
import mock
import pytest
from pytest_localserver.http import ContentServer
from werkzeug.wrappers import Request, Response

import elasticapm
from elasticapm.base import Client
from elasticapm.conf.constants import SPAN
from elasticapm.instrumentation import register
from elasticapm.traces import execution_context
from elasticapm.transport.http import Transport
from elasticapm.transport.http_base import HTTPTransportBase
from elasticapm.utils.threading import ThreadManager

cur_dir = os.path.dirname(os.path.realpath(__file__))

ERRORS_SCHEMA = os.path.join(cur_dir, "upstream", "json-specs", "error.json")
TRANSACTIONS_SCHEMA = os.path.join(cur_dir, "upstream", "json-specs", "transaction.json")
SPAN_SCHEMA = os.path.join(cur_dir, "upstream", "json-specs", "span.json")
METRICSET_SCHEMA = os.path.join(cur_dir, "upstream", "json-specs", "metricset.json")
METADATA_SCHEMA = os.path.join(cur_dir, "upstream", "json-specs", "metadata.json")

with open(os.path.join(cur_dir, "upstream", "json-specs", "span_types.json")) as f:
    SPAN_TYPES = json.load(f)


with open(ERRORS_SCHEMA, encoding="utf8") as errors_json, open(
    TRANSACTIONS_SCHEMA, encoding="utf8"
) as transactions_json, open(SPAN_SCHEMA, encoding="utf8") as span_json, open(
    METRICSET_SCHEMA, encoding="utf8"
) as metricset_json, open(
    METADATA_SCHEMA, encoding="utf8"
) as metadata_json:
    VALIDATORS = {
        "error": jsonschema.Draft4Validator(
            json.load(errors_json),
            resolver=jsonschema.RefResolver(
                base_uri="file:" + pathname2url(ERRORS_SCHEMA), referrer="file:" + pathname2url(ERRORS_SCHEMA)
            ),
        ),
        "transaction": jsonschema.Draft4Validator(
            json.load(transactions_json),
            resolver=jsonschema.RefResolver(
                base_uri="file:" + pathname2url(TRANSACTIONS_SCHEMA),
                referrer="file:" + pathname2url(TRANSACTIONS_SCHEMA),
            ),
        ),
        "span": jsonschema.Draft4Validator(
            json.load(span_json),
            resolver=jsonschema.RefResolver(
                base_uri="file:" + pathname2url(SPAN_SCHEMA), referrer="file:" + pathname2url(SPAN_SCHEMA)
            ),
        ),
        "metricset": jsonschema.Draft4Validator(
            json.load(metricset_json),
            resolver=jsonschema.RefResolver(
                base_uri="file:" + pathname2url(METRICSET_SCHEMA), referrer="file:" + pathname2url(METRICSET_SCHEMA)
            ),
        ),
        "metadata": jsonschema.Draft4Validator(
            json.load(metadata_json),
            resolver=jsonschema.RefResolver(
                base_uri="file:" + pathname2url(METADATA_SCHEMA), referrer="file:" + pathname2url(METADATA_SCHEMA)
            ),
        ),
    }


def validate_span_type_subtype(item: dict) -> Optional[str]:
    """
    Validate span type/subtype against spec.

    At first, only warnings are issued. At a later point, it should return the message as string
    which will cause a validation error.
    """
    if item["type"] not in SPAN_TYPES:
        warnings.warn(f"Span type \"{item['type']}\" not found in JSON spec", UserWarning)
        return
    span_type = SPAN_TYPES[item["type"]]
    subtypes = span_type.get("subtypes", [])
    if not subtypes and item["subtype"] and not span_type.get("allow_unlisted_subtype", False):
        warnings.warn(
            f"Span type \"{item['type']}\" has no subtypes, but subtype \"{item['subtype']}\" is set", UserWarning
        )
        return
    if item["subtype"] not in SPAN_TYPES[item["type"]].get("subtypes", []):
        if not SPAN_TYPES[item["type"]].get("allow_unlisted_subtype", False):
            warnings.warn(f"Subtype \"{item['subtype']}\" not allowed for span type \"{item['type']}\"", UserWarning)
            return
    else:
        if "python" not in subtypes.get(item["subtype"], {}).get("__used_by", []):
            warnings.warn(f"\"{item['type']}.{item['subtype']}\" not marked as used by Python", UserWarning)
    return None


class ValidatingWSGIApp(ContentServer):
    def __init__(self, **kwargs) -> None:
        self.skip_validate = kwargs.pop("skip_validate", False)
        super(ValidatingWSGIApp, self).__init__(**kwargs)
        self.payloads = []
        self.responses = []

    def __call__(self, environ, start_response):
        content = self.content
        request = Request(environ)
        self.requests.append(request)
        data = request.data
        if request.content_encoding == "deflate":
            data = zlib.decompress(data)
        elif request.content_encoding == "gzip":
            with gzip.GzipFile(fileobj=io.BytesIO(data)) as f:
                data = f.read()
        data = data.decode("utf-8")
        if request.content_type == "application/x-ndjson":
            data = [json.loads(line) for line in data.split("\n") if line]
        self.payloads.append(data)
        code = 202
        success = 0
        fail = 0
        if not self.skip_validate:
            for line in data:
                item_type, item = list(line.items())[0]
                validator = VALIDATORS[item_type]
                try:
                    validator.validate(item)
                    success += 1
                except jsonschema.ValidationError as e:
                    fail += 1
                    content += "/".join(map(str, e.absolute_schema_path)) + ": " + e.message + "\n"
                if item_type == "span":
                    result = validate_span_type_subtype(item)
                    if result:
                        fail += 1
                        content += result
            code = 202 if not fail else 400
        response = Response(status=code)
        response.headers.clear()
        response.headers.extend(self.headers)
        response.data = content
        self.responses.append({"code": code, "content": content})
        return response(environ, start_response)


@pytest.fixture
def mock_client_excepthook():
    with mock.patch("tests.fixtures.TempStoreClient._excepthook") as m:
        yield m


@pytest.fixture
def mock_client_capture_exception():
    with mock.patch("tests.fixtures.TempStoreClient.capture_exception") as m:
        yield m


@pytest.fixture
def original_exception_hook(request):
    mock_params = getattr(request, "param", {})
    original_excepthook = sys.excepthook
    mck = mock.Mock(side_effect=mock_params.get("side_effect"))
    sys.excepthook = mck
    yield mck
    sys.excepthook = original_excepthook


@pytest.fixture()
def elasticapm_client(request):
    original_exceptionhook = sys.excepthook
    client_config = getattr(request, "param", {})
    client_class = client_config.pop("client_class", TempStoreClient)
    client_config.setdefault("service_name", "myapp")
    client_config.setdefault("secret_token", "test_key")
    client_config.setdefault("central_config", "false")
    client_config.setdefault("include_paths", ("*/tests/*",))
    client_config.setdefault("span_stack_trace_min_duration", 0)
    client_config.setdefault("metrics_interval", "0ms")
    client_config.setdefault("cloud_provider", False)
    client_config.setdefault("span_compression_exact_match_max_duration", "0ms")
    client_config.setdefault("span_compression_same_kind_max_duration", "0ms")
    client_config.setdefault("exit_span_min_duration", "0ms")
    client = client_class(**client_config)
    yield client
    client.close()
    # clear any execution context that might linger around
    sys.excepthook = original_exceptionhook
    execution_context.set_transaction(None)
    execution_context.unset_span(clear_all=True)
    if client._transport.validation_errors:
        pytest.fail(
            "Validation errors:" + "\n".join(*itertools.chain(v for v in client._transport.validation_errors.values()))
        )
    logger = logging.getLogger("elasticapm")
    logger.setLevel(logging.NOTSET)


@pytest.fixture()
def elasticapm_transaction(elasticapm_client):
    """
    Useful fixture if spans from other fixtures should be captured.
    This can be achieved by listing this fixture first.
    """
    transaction = elasticapm_client.begin_transaction("test")
    yield transaction


@pytest.fixture()
def elasticapm_client_log_file(request):
    original_exceptionhook = sys.excepthook
    client_config = getattr(request, "param", {})
    client_config.setdefault("service_name", "myapp")
    client_config.setdefault("secret_token", "test_key")
    client_config.setdefault("central_config", "false")
    client_config.setdefault("include_paths", ("*/tests/*",))
    client_config.setdefault("span_stack_trace_min_duration", 0)
    client_config.setdefault("span_compression_exact_match_max_duration", "0ms")
    client_config.setdefault("span_compression_same_kind_max_duration", "0ms")
    client_config.setdefault("metrics_interval", "0ms")
    client_config.setdefault("cloud_provider", False)
    client_config.setdefault("log_level", "warning")

    root_logger = logging.getLogger()
    handler = logging.StreamHandler()
    root_logger.addHandler(handler)

    tmp = tempfile.NamedTemporaryFile(delete=False)
    tmp.close()
    client_config["log_file"] = tmp.name

    client = TempStoreClient(**client_config)
    yield client
    client.close()

    # delete our tmpfile
    logger = logging.getLogger("elasticapm")
    for handler in logger.handlers:
        if isinstance(handler, logging.handlers.RotatingFileHandler):
            handler.close()
    os.unlink(tmp.name)

    # Remove our streamhandler
    root_logger.removeHandler(handler)

    # clear any execution context that might linger around
    sys.excepthook = original_exceptionhook
    execution_context.set_transaction(None)
    execution_context.unset_span(clear_all=True)
    logger = logging.getLogger("elasticapm")
    logger.setLevel(logging.NOTSET)


@pytest.fixture()
def waiting_httpserver(httpserver):
    wait_for_open_port(httpserver.server_address[1])
    return httpserver


@pytest.fixture
def httpsserver_custom(request):
    """The returned ``httpsserver`` (note the additional S!) provides a
    threaded HTTP server instance similar to funcarg ``httpserver`` but with
    SSL encryption.
    """
    from pytest_localserver import https

    config = getattr(request, "param", {})
    key = os.path.join(cur_dir, "ca", config.get("key", "server.pem"))

    server = https.SecureContentServer(key=key, cert=key)
    server.start()
    request.addfinalizer(server.stop)
    return server


@pytest.fixture()
def waiting_httpsserver(httpsserver_custom):
    wait_for_open_port(httpsserver_custom.server_address[1])
    return httpsserver_custom


@pytest.fixture()
def validating_httpserver(request):
    config = getattr(request, "param", {})
    app = config.pop("app", ValidatingWSGIApp)
    server = app(**config)
    server.start()
    wait_for_open_port(server.server_address[1])
    request.addfinalizer(server.stop)
    return server


@pytest.fixture()
def sending_elasticapm_client(request, validating_httpserver):
    validating_httpserver.serve_content(code=202, content="", headers={"Location": "http://example.com/foo"})
    client_config = getattr(request, "param", {})
    client_config.setdefault("server_url", validating_httpserver.url)
    client_config.setdefault("service_name", "myapp")
    client_config.setdefault("secret_token", "test_key")
    client_config.setdefault("transport_class", "elasticapm.transport.http.Transport")
    client_config.setdefault("span_stack_trace_min_duration", 0)
    client_config.setdefault("span_compression_exact_match_max_duration", "0ms")
    client_config.setdefault("span_compression_same_kind_max_duration", "0ms")
    client_config.setdefault("include_paths", ("*/tests/*",))
    client_config.setdefault("metrics_interval", "0ms")
    client_config.setdefault("cloud_provider", False)
    client_config.setdefault("central_config", "false")
    client_config.setdefault("server_version", (8, 0, 0))
    client = Client(**client_config)
    client.httpserver = validating_httpserver
    yield client
    client.close()
    # clear any execution context that might linger around
    execution_context.set_transaction(None)
    execution_context.unset_span(clear_all=True)


class DummyTransport(HTTPTransportBase):
    def __init__(self, url, *args, **kwargs) -> None:
        super(DummyTransport, self).__init__(url, *args, **kwargs)
        self.events = defaultdict(list)
        self.validation_errors = defaultdict(list)

    def queue(self, event_type, data, flush=False) -> None:
        self._flushed.clear()
        data = self._process_event(event_type, data)
        self.events[event_type].append(data)
        self._flushed.set()
        if data is not None:
            validator = VALIDATORS[event_type]
            try:
                validator.validate(data)
            except jsonschema.ValidationError as e:
                self.validation_errors[event_type].append(e.message)
            if event_type == "span":
                result = validate_span_type_subtype(data)
                if result:
                    self.validation_errors[event_type].append(result)

    def start_thread(self, pid=None) -> None:
        # don't call the parent method, but the one from ThreadManager
        ThreadManager.start_thread(self, pid=pid)

    def stop_thread(self) -> None:
        pass

    def get_config(self, current_version=None, keys=None):
        return False, None, 30


class MockSendHTTPTransport(Transport):
    """Mocking the send method of the Transport class sometimes fails silently in client tests.
    After spending some time trying to understand this with no luck just use this class instead."""

    def __init__(self, url, *args, **kwargs):
        self.send_mock = mock.Mock()
        super().__init__(url, *args, **kwargs)

    def send(self, data, forced_flush=False, custom_url=None, custom_headers=None):
        return self.send_mock(data, forced_flush, custom_url, custom_headers)


class TempStoreClient(Client):
    def __init__(self, config=None, **inline) -> None:
        inline.setdefault("transport_class", "tests.fixtures.DummyTransport")
        super(TempStoreClient, self).__init__(config, **inline)

    @property
    def events(self):
        return self._transport.events

    def spans_for_transaction(self, transaction):
        """Test helper method to get all spans of a specific transaction"""
        return [span for span in self.events[SPAN] if span["transaction_id"] == transaction["id"]]


@pytest.fixture()
def temp_store_client():
    return TempStoreClient


@pytest.fixture()
def not_so_random():
    old_state = random.getstate()
    random.seed(42)
    yield
    random.setstate(old_state)


@pytest.fixture()
def instrument():
    elasticapm.instrument()
    yield
    elasticapm.uninstrument()


@pytest.fixture()
def wrapper_instrument():
    old_register = register._cls_register.copy()
    register.register_wrapper_instrumentations()
    elasticapm.instrument()
    yield
    elasticapm.uninstrument()
    register._cls_register = old_register


def wait_for_open_port(port: int, host: str = "localhost", timeout: int = 30):
    start_time = time.time()
    while True:
        try:
            sock = socket.create_connection((host, port), timeout=0.1)
            sock.close()
            break
        except socket.error:
            time.sleep(0.01)
            if time.time() - start_time > timeout:
                raise TimeoutError()


def get_free_port() -> int:
    with socketserver.TCPServer(("localhost", 0), None) as s:
        free_port = s.server_address[1]
    return free_port


@pytest.fixture(autouse=True)
def always_uninstrument_and_close():
    """
    It's easy to accidentally forget to uninstrument, or to leave a Client open.

    With no-code-changes instrumentations, we *really* need to make sure we
    always uninstrument. This fixture will be used on every test, should be
    applied first -- see
    https://docs.pytest.org/en/stable/reference/fixtures.html#autouse-fixtures-are-executed-first-within-their-scope
    -- and thus cleanup last, which will ensure we always uninstrument and close
    the Client.
    """
    try:
        yield
    finally:
        try:
            elasticapm.uninstrument()
            client = elasticapm.get_client()
            if client:
                client.close()
        except Exception:
            pass


@pytest.fixture()
def invalidate_fqdn_cache():
    fqdn = elasticapm.utils.fqdn
    elasticapm.utils.fqdn = None
    yield
    elasticapm.utils.fqdn = fqdn



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from . import testgrpc_pb2 as test__pb2


class TestServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.GetServerResponse = channel.unary_unary(
                '/test.TestService/GetServerResponse',
                request_serializer=test__pb2.Message.SerializeToString,
                response_deserializer=test__pb2.MessageResponse.FromString,
                )
        self.GetServerResponseAbort = channel.unary_unary(
                '/test.TestService/GetServerResponseAbort',
                request_serializer=test__pb2.Message.SerializeToString,
                response_deserializer=test__pb2.MessageResponse.FromString,
                )
        self.GetServerResponseUnavailable = channel.unary_unary(
                '/test.TestService/GetServerResponseUnavailable',
                request_serializer=test__pb2.Message.SerializeToString,
                response_deserializer=test__pb2.MessageResponse.FromString,
                )
        self.GetServerResponseException = channel.unary_unary(
                '/test.TestService/GetServerResponseException',
                request_serializer=test__pb2.Message.SerializeToString,
                response_deserializer=test__pb2.MessageResponse.FromString,
                )


class TestServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def GetServerResponse(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetServerResponseAbort(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetServerResponseUnavailable(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetServerResponseException(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_TestServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'GetServerResponse': grpc.unary_unary_rpc_method_handler(
                    servicer.GetServerResponse,
                    request_deserializer=test__pb2.Message.FromString,
                    response_serializer=test__pb2.MessageResponse.SerializeToString,
            ),
            'GetServerResponseAbort': grpc.unary_unary_rpc_method_handler(
                    servicer.GetServerResponseAbort,
                    request_deserializer=test__pb2.Message.FromString,
                    response_serializer=test__pb2.MessageResponse.SerializeToString,
            ),
            'GetServerResponseUnavailable': grpc.unary_unary_rpc_method_handler(
                    servicer.GetServerResponseUnavailable,
                    request_deserializer=test__pb2.Message.FromString,
                    response_serializer=test__pb2.MessageResponse.SerializeToString,
            ),
            'GetServerResponseException': grpc.unary_unary_rpc_method_handler(
                    servicer.GetServerResponseException,
                    request_deserializer=test__pb2.Message.FromString,
                    response_serializer=test__pb2.MessageResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'test.TestService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class TestService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def GetServerResponse(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/test.TestService/GetServerResponse',
            test__pb2.Message.SerializeToString,
            test__pb2.MessageResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetServerResponseAbort(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/test.TestService/GetServerResponseAbort',
            test__pb2.Message.SerializeToString,
            test__pb2.MessageResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetServerResponseUnavailable(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/test.TestService/GetServerResponseUnavailable',
            test__pb2.Message.SerializeToString,
            test__pb2.MessageResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetServerResponseException(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/test.TestService/GetServerResponseException',
            test__pb2.Message.SerializeToString,
            test__pb2.MessageResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import urllib.parse
from urllib.error import HTTPError, URLError
from urllib.request import urlopen

import mock
import pytest

from elasticapm.conf import constants
from elasticapm.conf.constants import SPAN, TRANSACTION
from elasticapm.traces import capture_span
from elasticapm.utils.disttracing import TraceParent

request_method = "http.client.HTTPConnection.request"
getresponse_method = "http.client.HTTPConnection.getresponse"


def test_urllib(instrument, elasticapm_client, waiting_httpserver):
    waiting_httpserver.serve_content("")
    url = waiting_httpserver.url + "/hello_world"
    parsed_url = urllib.parse.urlparse(url)
    elasticapm_client.begin_transaction("transaction")
    expected_sig = "GET {0}".format(parsed_url.netloc)
    with capture_span("test_name", "test_type"):
        url = "http://{0}/hello_world".format(parsed_url.netloc)
        r = urlopen(url)

    elasticapm_client.end_transaction("MyView")

    transactions = elasticapm_client.events[TRANSACTION]
    spans = elasticapm_client.spans_for_transaction(transactions[0])

    expected_signatures = {"test_name", expected_sig}

    assert {t["name"] for t in spans} == expected_signatures

    assert len(spans) == 2

    assert spans[0]["name"] == expected_sig
    assert spans[0]["type"] == "external"
    assert spans[0]["subtype"] == "http"
    assert spans[0]["context"]["http"]["url"] == url
    assert spans[0]["context"]["http"]["status_code"] == 200
    assert spans[0]["context"]["destination"]["service"] == {
        "name": "",
        "resource": "127.0.0.1:%d" % parsed_url.port,
        "type": "",
    }
    assert spans[0]["context"]["service"]["target"]["type"] == "http"
    assert spans[0]["context"]["service"]["target"]["name"] == f"127.0.0.1:{parsed_url.port}"
    assert spans[0]["parent_id"] == spans[1]["id"]
    assert spans[0]["outcome"] == "success"

    assert spans[1]["name"] == "test_name"
    assert spans[1]["type"] == "test_type"
    assert spans[1]["parent_id"] == transactions[0]["id"]


@pytest.mark.parametrize("status_code", [400, 500])
def test_urllib_error(instrument, elasticapm_client, waiting_httpserver, status_code):
    waiting_httpserver.serve_content("", code=status_code)
    url = waiting_httpserver.url + "/hello_world"
    parsed_url = urllib.parse.urlparse(url)
    elasticapm_client.begin_transaction("transaction")
    expected_sig = "GET {0}".format(parsed_url.netloc)
    with capture_span("test_name", "test_type"):
        url = "http://{0}/hello_world".format(parsed_url.netloc)
        try:
            r = urlopen(url)
        except HTTPError:
            pass

    elasticapm_client.end_transaction("MyView")

    transactions = elasticapm_client.events[TRANSACTION]
    spans = elasticapm_client.spans_for_transaction(transactions[0])

    assert spans[0]["name"] == expected_sig
    assert spans[0]["type"] == "external"
    assert spans[0]["subtype"] == "http"
    assert spans[0]["context"]["http"]["url"] == url
    assert spans[0]["context"]["http"]["status_code"] == status_code
    assert spans[0]["outcome"] == "failure"


@mock.patch(request_method)
@mock.patch(getresponse_method)
def test_urllib_standard_port(mock_getresponse, mock_request, instrument, elasticapm_client):
    # Python internally used both "code" and "status"
    mock_getresponse.return_value = mock.Mock(code=200, status=200)

    url = "http://example.com/"
    parsed_url = urllib.parse.urlparse(url)
    elasticapm_client.begin_transaction("transaction")
    expected_sig = "GET {0}".format(parsed_url.netloc)
    r = urlopen(url)

    elasticapm_client.end_transaction("MyView")

    transactions = elasticapm_client.events[TRANSACTION]
    spans = elasticapm_client.spans_for_transaction(transactions[0])

    assert spans[0]["name"] == expected_sig
    assert spans[0]["type"] == "external"
    assert spans[0]["subtype"] == "http"
    assert spans[0]["context"]["http"]["url"] == url
    assert spans[0]["parent_id"] == transactions[0]["id"]


@pytest.mark.parametrize(
    "elasticapm_client",
    [
        pytest.param({"use_elastic_traceparent_header": True}, id="use_elastic_traceparent_header-True"),
        pytest.param({"use_elastic_traceparent_header": False}, id="use_elastic_traceparent_header-False"),
    ],
    indirect=True,
)
def test_trace_parent_propagation_sampled(instrument, elasticapm_client, waiting_httpserver):
    waiting_httpserver.serve_content("")
    url = waiting_httpserver.url + "/hello_world"
    elasticapm_client.begin_transaction("transaction")
    urlopen(url)
    elasticapm_client.end_transaction("MyView")
    transactions = elasticapm_client.events[TRANSACTION]
    spans = elasticapm_client.spans_for_transaction(transactions[0])

    headers = waiting_httpserver.requests[0].headers
    assert constants.TRACEPARENT_HEADER_NAME in headers
    trace_parent = TraceParent.from_string(
        headers[constants.TRACEPARENT_HEADER_NAME], tracestate_string=headers[constants.TRACESTATE_HEADER_NAME]
    )
    assert trace_parent.trace_id == transactions[0]["trace_id"]
    assert trace_parent.span_id == spans[0]["id"]
    assert trace_parent.trace_options.recorded
    # Check that sample_rate was correctly placed in the tracestate
    assert constants.TRACESTATE.SAMPLE_RATE in trace_parent.tracestate_dict

    if elasticapm_client.config.use_elastic_traceparent_header:
        assert constants.TRACEPARENT_LEGACY_HEADER_NAME in headers
        assert headers[constants.TRACEPARENT_HEADER_NAME] == headers[constants.TRACEPARENT_LEGACY_HEADER_NAME]
    else:
        assert constants.TRACEPARENT_LEGACY_HEADER_NAME not in headers


@pytest.mark.parametrize(
    "elasticapm_client",
    [
        pytest.param({"use_elastic_traceparent_header": True}, id="use_elastic_traceparent_header-True"),
        pytest.param({"use_elastic_traceparent_header": False}, id="use_elastic_traceparent_header-False"),
    ],
    indirect=True,
)
def test_trace_parent_propagation_unsampled(instrument, elasticapm_client, waiting_httpserver):
    waiting_httpserver.serve_content("")
    url = waiting_httpserver.url + "/hello_world"
    transaction_object = elasticapm_client.begin_transaction("transaction")
    transaction_object.is_sampled = False
    urlopen(url)
    elasticapm_client.end_transaction("MyView")
    spans = elasticapm_client.events[SPAN]

    assert not spans

    headers = waiting_httpserver.requests[0].headers
    assert constants.TRACEPARENT_HEADER_NAME in headers
    trace_parent = TraceParent.from_string(
        headers[constants.TRACEPARENT_HEADER_NAME], tracestate_string=headers[constants.TRACESTATE_HEADER_NAME]
    )
    assert trace_parent.trace_id == transaction_object.trace_parent.trace_id
    assert trace_parent.span_id == transaction_object.id
    assert not trace_parent.trace_options.recorded
    # Check that sample_rate was correctly placed in the tracestate
    assert constants.TRACESTATE.SAMPLE_RATE in trace_parent.tracestate_dict

    if elasticapm_client.config.use_elastic_traceparent_header:
        assert constants.TRACEPARENT_LEGACY_HEADER_NAME in headers
        assert headers[constants.TRACEPARENT_HEADER_NAME] == headers[constants.TRACEPARENT_LEGACY_HEADER_NAME]
    else:
        assert constants.TRACEPARENT_LEGACY_HEADER_NAME not in headers


@pytest.mark.parametrize(
    "is_sampled", [pytest.param(True, id="is_sampled-True"), pytest.param(False, id="is_sampled-False")]
)
def test_tracestate_propagation(instrument, elasticapm_client, waiting_httpserver, is_sampled):
    traceparent = TraceParent.from_string(
        "00-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-03", "foo=bar,baz=bazzinga"
    )

    waiting_httpserver.serve_content("")
    url = waiting_httpserver.url + "/hello_world"
    transaction_object = elasticapm_client.begin_transaction("transaction", trace_parent=traceparent)
    transaction_object.is_sampled = is_sampled
    urlopen(url)
    elasticapm_client.end_transaction("MyView")
    headers = waiting_httpserver.requests[0].headers
    assert headers[constants.TRACESTATE_HEADER_NAME] == "foo=bar,baz=bazzinga"


@pytest.mark.parametrize("elasticapm_client", [{"transaction_max_spans": 1}], indirect=True)
def test_span_only_dropped(instrument, elasticapm_client, waiting_httpserver):
    """test that urllib instrumentation does not fail if no parent span can be found"""
    waiting_httpserver.serve_content("")
    url = waiting_httpserver.url + "/hello_world"
    transaction_object = elasticapm_client.begin_transaction("transaction")
    for i in range(2):
        with capture_span("test", "test"):
            urlopen(url)
    elasticapm_client.end_transaction("bla", "OK")
    trace_parent_1 = TraceParent.from_string(waiting_httpserver.requests[0].headers[constants.TRACEPARENT_HEADER_NAME])
    trace_parent_2 = TraceParent.from_string(waiting_httpserver.requests[1].headers[constants.TRACEPARENT_HEADER_NAME])

    assert trace_parent_1.span_id != transaction_object.id
    # second request should use transaction id as span id because there is no span
    assert trace_parent_2.span_id == transaction_object.id


def test_url_sanitization(instrument, elasticapm_client, waiting_httpserver):
    waiting_httpserver.serve_content("")
    url = waiting_httpserver.url + "/hello_world"
    url = url.replace("http://", "http://user:pass@")
    transaction_object = elasticapm_client.begin_transaction("transaction")
    with pytest.raises(URLError):
        urlopen(url)
    elasticapm_client.end_transaction("MyView")
    transactions = elasticapm_client.events[TRANSACTION]
    span = elasticapm_client.spans_for_transaction(transactions[0])[0]

    assert "pass" not in span["context"]["http"]["url"]
    assert constants.MASK_URL in span["context"]["http"]["url"]



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from __future__ import absolute_import

from elasticapm.contrib.asyncio.traces import async_capture_span
from elasticapm.instrumentation.packages.asyncio.dbapi2_asyncio import (
    AsyncConnectionProxy,
    AsyncCursorProxy,
    AsyncDbApi2Instrumentation,
)
from elasticapm.instrumentation.packages.dbapi2 import extract_signature
from elasticapm.instrumentation.packages.psycopg2 import get_destination_info


class PGAsyncCursorProxy(AsyncCursorProxy):
    provider_name = "postgresql"

    def _bake_sql(self, sql):
        # If this is a Composable object, use its `as_string` method.
        # See https://www.psycopg.org/psycopg3/docs/api/sql.html
        if hasattr(sql, "as_string"):
            sql = sql.as_string(self.__wrapped__)
        # If the sql string is already a byte string, we need to decode it using the connection encoding
        if isinstance(sql, bytes):
            sql = sql.decode(self.connection.info.encoding)
        return sql

    def extract_signature(self, sql):
        return extract_signature(sql)

    async def execute(self, query, params=None, **kwargs):
        return await self._trace_sql(self.__wrapped__.execute, query, params, **kwargs)

    async def executemany(self, query, params_seq, **kwargs):
        return await self._trace_sql(self.__wrapped__.executemany, query, params_seq, **kwargs)

    async def __aenter__(self):
        return PGAsyncCursorProxy(await self.__wrapped__.__aenter__(), destination_info=self._self_destination_info)

    async def __aexit__(self, *args):
        return PGAsyncCursorProxy(await self.__wrapped__.__aexit__(*args), destination_info=self._self_destination_info)

    def __aiter__(self):
        return self.__wrapped__.__aiter__()

    @property
    def _self_database(self):
        return self.connection.info.dbname or ""


class PGAsyncConnectionProxy(AsyncConnectionProxy):
    cursor_proxy = PGAsyncCursorProxy

    async def __aenter__(self):
        return PGAsyncConnectionProxy(await self.__wrapped__.__aenter__(), destination_info=self._self_destination_info)

    async def __aexit__(self, *args):
        return PGAsyncConnectionProxy(
            await self.__wrapped__.__aexit__(*args), destination_info=self._self_destination_info
        )


class AsyncPsycopgInstrumentation(AsyncDbApi2Instrumentation):
    name = "psycopg_async"

    instrument_list = [("psycopg.connection_async", "AsyncConnection.connect")]

    async def call(self, module, method, wrapped, instance, args, kwargs):
        signature = "psycopg.connect_async"

        host, port = get_destination_info(kwargs.get("host"), kwargs.get("port"))
        database = kwargs.get("dbname")
        signature = f"{signature} {host}:{port}"  # noqa: E231
        destination_info = {
            "address": host,
            "port": port,
        }
        async with async_capture_span(
            signature,
            span_type="db",
            span_subtype="postgresql",
            span_action="connect",
            leaf=True,
            extra={"destination": destination_info, "db": {"type": "sql", "instance": database}},
        ):
            return PGAsyncConnectionProxy(await wrapped(*args, **kwargs), destination_info=destination_info)



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import logging
import time
from multiprocessing.dummy import Pool

import mock
import pytest

from elasticapm.conf import constants
from elasticapm.metrics.base_metrics import Counter, Gauge, MetricSet, MetricsRegistry, NoopMetric, Timer
from tests.utils import assert_any_record_contains


class DummyMetricSet(MetricSet):
    def before_collect(self):
        self.gauge("a.b.c.d").val = 0
        self.gauge("a").val = 0
        self.gauge("b").val = 0
        self.gauge("c").val = 0


@pytest.mark.parametrize("elasticapm_client", [{"metrics_interval": "30s"}], indirect=True)
def test_metrics_registry(elasticapm_client):
    registry = MetricsRegistry(elasticapm_client)
    registry.register("tests.metrics.base_tests.DummyMetricSet")
    registry.collect()
    assert len(elasticapm_client.events[constants.METRICSET])


@pytest.mark.parametrize("elasticapm_client", [{"metrics_interval": "30s"}], indirect=True)
def test_metrics_registry_instance(elasticapm_client):
    registry = MetricsRegistry(elasticapm_client)
    registry.register(DummyMetricSet)
    registry.collect()
    assert len(elasticapm_client.events[constants.METRICSET])


@pytest.mark.parametrize(
    "elasticapm_client",
    [{"metrics_sets": "tests.metrics.base_tests.DummyMetricSet", "disable_metrics": "a.*,*c"}],
    indirect=True,
)
def test_disable_metrics(elasticapm_client):
    elasticapm_client.metrics.collect()
    metrics = elasticapm_client.events[constants.METRICSET][0]
    assert "a" in metrics["samples"]
    assert "b" in metrics["samples"]
    assert "a.b.c.d" not in metrics["samples"]
    assert "c" not in metrics["samples"]


def test_metrics_counter(elasticapm_client):
    metricset = MetricSet(MetricsRegistry(elasticapm_client))
    metricset.counter("x").inc()
    data = next(metricset.collect())
    assert data["samples"]["x"]["value"] == 1
    metricset.counter("x").inc(10)
    data = next(metricset.collect())
    assert data["samples"]["x"]["value"] == 11
    metricset.counter("x").dec(10)
    data = next(metricset.collect())
    assert data["samples"]["x"]["value"] == 1
    metricset.counter("x").dec()
    data = next(metricset.collect())
    assert data["samples"]["x"]["value"] == 0


def test_metrics_histogram(elasticapm_client):
    metricset = MetricSet(MetricsRegistry(elasticapm_client))
    hist = metricset.histogram("x", buckets=[1, 10, 100])
    assert len(hist.buckets) == 4

    hist.update(0.3)
    hist.update(1)
    hist.update(5)
    hist.update(20)
    hist.update(100)
    hist.update(1000)

    data = list(metricset.collect())
    assert len(data) == 1
    d = data[0]
    assert d["samples"]["x"]["counts"] == [2, 1, 2, 1]
    assert d["samples"]["x"]["values"] == [0.5, 5.5, 55.0, 100]


def test_metrics_labels(elasticapm_client):
    metricset = MetricSet(MetricsRegistry(elasticapm_client))
    metricset.counter("x", mylabel="a").inc()
    metricset.counter("y", mylabel="a").inc()
    metricset.counter("x", mylabel="b").inc().inc()
    metricset.counter("x", mylabel="b", myotherlabel="c").inc()
    metricset.counter("x", mylabel="a").dec()
    data = list(metricset.collect())
    asserts = 0
    for d in data:
        if d["tags"] == {"mylabel": "a"}:
            assert d["samples"]["x"]["value"] == 0
            assert d["samples"]["y"]["value"] == 1
            asserts += 1
        elif d["tags"] == {"mylabel": "b"}:
            assert d["samples"]["x"]["value"] == 2
            asserts += 1
        elif d["tags"] == {"mylabel": "b", "myotherlabel": "c"}:
            assert d["samples"]["x"]["value"] == 1
            asserts += 1
    assert asserts == 3


def test_metrics_multithreaded(elasticapm_client):
    metricset = MetricSet(MetricsRegistry(elasticapm_client))
    pool = Pool(5)

    def target():
        for i in range(500):
            metricset.counter("x").inc(i + 1)
            time.sleep(0.0000001)

    [pool.apply_async(target, ()) for i in range(10)]
    pool.close()
    pool.join()
    expected = 10 * ((500 * 501) / 2)
    assert metricset.counter("x").val == expected


@pytest.mark.parametrize("sending_elasticapm_client", [{"metrics_interval": "30s"}], indirect=True)
def test_metrics_flushed_on_shutdown(sending_elasticapm_client):
    # this is ugly, we need an API for this at some point...
    metricset = MetricSet(sending_elasticapm_client.metrics)
    sending_elasticapm_client.metrics._metricsets["foo"] = metricset
    metricset.counter("x").inc()
    sending_elasticapm_client.close()
    assert sending_elasticapm_client.httpserver.payloads
    for item in sending_elasticapm_client.httpserver.payloads[0]:
        try:
            assert item["metricset"]["samples"]["x"]["value"] == 1
            break
        except KeyError:
            pass
    else:
        assert False, "no item found with matching dict path metricset.samples.x.value"


@mock.patch("elasticapm.metrics.base_metrics.DISTINCT_LABEL_LIMIT", 3)
def test_metric_limit(caplog, elasticapm_client):
    m = MetricSet(MetricsRegistry(elasticapm_client))
    with caplog.at_level(logging.WARNING, logger="elasticapm.metrics"):
        for i in range(2):
            counter = m.counter("counter", some_label=i)
            gauge = m.gauge("gauge", some_label=i)
            timer = m.timer("timer", some_label=i)
            if i == 0:
                assert isinstance(timer, Timer)
                assert isinstance(gauge, Gauge)
                assert isinstance(counter, Counter)
            else:
                assert isinstance(timer, NoopMetric)
                assert isinstance(gauge, NoopMetric)
                assert isinstance(counter, NoopMetric)
    assert_any_record_contains(caplog.records, "The limit of 3 metricsets has been reached", "elasticapm.metrics")


def test_metrics_not_collected_if_zero_and_reset(elasticapm_client):
    m = MetricSet(MetricsRegistry(elasticapm_client))
    counter = m.counter("counter", reset_on_collect=False)
    resetting_counter = m.counter("resetting_counter", reset_on_collect=True)
    gauge = m.gauge("gauge", reset_on_collect=False)
    resetting_gauge = m.gauge("resetting_gauge", reset_on_collect=True)
    timer = m.timer("timer", reset_on_collect=False, unit="us")
    resetting_timer = m.timer("resetting_timer", reset_on_collect=True, unit="us")

    counter.inc(), resetting_counter.inc()
    gauge.val = 5
    resetting_gauge.val = 5
    timer.update(1, 1)
    resetting_timer.update(1, 1)

    data = list(m.collect())
    more_data = list(m.collect())
    assert set(data[0]["samples"].keys()) == {
        "counter",
        "resetting_counter",
        "gauge",
        "resetting_gauge",
        "timer.count",
        "timer.sum.us",
        "resetting_timer.count",
        "resetting_timer.sum.us",
    }
    assert set(more_data[0]["samples"].keys()) == {"counter", "gauge", "timer.count", "timer.sum.us"}


def test_underscore_metrics_deprecation(elasticapm_client):
    with pytest.warns(DeprecationWarning):
        elasticapm_client._metrics



#  BSD 3-Clause License
#
#  Copyright (c) 2012, the Sentry Team, see AUTHORS for more details
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE


import warnings
from collections import defaultdict

from elasticapm.conf.constants import BASE_SANITIZE_FIELD_NAMES, ERROR, MASK, SPAN, TRANSACTION
from elasticapm.utils import varmap
from elasticapm.utils.encoding import force_text
from elasticapm.utils.stacks import get_lines_from_file


def for_events(*events):
    """
    :param events: list of event types
    Only calls wrapped function if given event_type is in list of events
    """
    events = set(events)

    def wrap(func):
        func.event_types = events
        return func

    return wrap


@for_events(ERROR, TRANSACTION)
def remove_http_request_body(client, event):
    """
    Removes request.body from context
    :param client: an ElasticAPM client
    :param event: a transaction or error event
    :return: The modified event
    """
    if "context" in event and "request" in event["context"]:
        event["context"]["request"].pop("body", None)
    return event


@for_events(ERROR, SPAN)
def remove_stacktrace_locals(client, event):
    """
    Removes local variables from any frames.
    :param client: an ElasticAPM client
    :param event: a transaction or error event
    :return: The modified event
    """
    func = lambda frame: frame.pop("vars", None)
    return _process_stack_frames(event, func)


@for_events(ERROR, SPAN)
def sanitize_stacktrace_locals(client, event):
    """
    Sanitizes local variables in all frames
    :param client: an ElasticAPM client
    :param event: a transaction or error event
    :return: The modified event
    """

    def func(frame) -> None:
        if "vars" in frame:
            frame["vars"] = varmap(_sanitize, frame["vars"], sanitize_field_names=client.config.sanitize_field_names)

    return _process_stack_frames(event, func)


@for_events(ERROR, TRANSACTION)
def sanitize_http_request_cookies(client, event):
    """
    Sanitizes http request cookies
    :param client: an ElasticAPM client
    :param event: a transaction or error event
    :return: The modified event
    """

    # sanitize request.cookies dict
    try:
        cookies = event["context"]["request"]["cookies"]
        event["context"]["request"]["cookies"] = varmap(
            _sanitize, cookies, sanitize_field_names=client.config.sanitize_field_names
        )
    except (KeyError, TypeError):
        pass

    # sanitize request.header.cookie string
    try:
        cookie_string = force_text(event["context"]["request"]["headers"]["cookie"], errors="replace")
        event["context"]["request"]["headers"]["cookie"] = _sanitize_string(
            cookie_string, "; ", "=", sanitize_field_names=client.config.sanitize_field_names
        )
    except (KeyError, TypeError):
        try:
            # Sometimes it's Cookie, not cookie
            cookie_string = force_text(event["context"]["request"]["headers"]["Cookie"], errors="replace")
            event["context"]["request"]["headers"]["Cookie"] = _sanitize_string(
                cookie_string, "; ", "=", sanitize_field_names=client.config.sanitize_field_names
            )
        except (KeyError, TypeError):
            pass
    return event


@for_events(ERROR, TRANSACTION)
def sanitize_http_response_cookies(client, event):
    """
    Sanitizes the set-cookie header of the response
    :param client: an ElasticAPM client
    :param event: a transaction or error event
    :return: The modified event
    """
    try:
        cookie_string = force_text(event["context"]["response"]["headers"]["set-cookie"], errors="replace")
        event["context"]["response"]["headers"]["set-cookie"] = _sanitize_string(
            cookie_string, ";", "=", sanitize_field_names=client.config.sanitize_field_names
        )
    except (KeyError, TypeError):
        pass
    return event


@for_events(ERROR, TRANSACTION)
def sanitize_http_headers(client, event):
    """
    Sanitizes http request/response headers
    :param client: an ElasticAPM client
    :param event: a transaction or error event
    :return: The modified event
    """
    # request headers
    try:
        headers = event["context"]["request"]["headers"]
        event["context"]["request"]["headers"] = varmap(
            _sanitize, headers, sanitize_field_names=client.config.sanitize_field_names
        )
    except (KeyError, TypeError):
        pass

    # response headers
    try:
        headers = event["context"]["response"]["headers"]
        event["context"]["response"]["headers"] = varmap(
            _sanitize, headers, sanitize_field_names=client.config.sanitize_field_names
        )
    except (KeyError, TypeError):
        pass

    return event


@for_events(ERROR, TRANSACTION)
def sanitize_http_wsgi_env(client, event):
    """
    Sanitizes WSGI environment variables
    :param client: an ElasticAPM client
    :param event: a transaction or error event
    :return: The modified event
    """
    try:
        env = event["context"]["request"]["env"]
        event["context"]["request"]["env"] = varmap(
            _sanitize, env, sanitize_field_names=client.config.sanitize_field_names
        )
    except (KeyError, TypeError):
        pass
    return event


@for_events(ERROR, TRANSACTION)
def sanitize_http_request_body(client, event):
    """
    Sanitizes http request body. This only works if the request body
    is a query-encoded string. Other types (e.g. JSON) are not handled by
    this sanitizer.
    :param client: an ElasticAPM client
    :param event: a transaction or error event
    :return: The modified event
    """
    try:
        body = force_text(event["context"]["request"]["body"], errors="replace")
    except (KeyError, TypeError):
        return event
    if "=" in body:
        sanitized_query_string = _sanitize_string(
            body, "&", "=", sanitize_field_names=client.config.sanitize_field_names
        )
        event["context"]["request"]["body"] = sanitized_query_string
    return event


@for_events(ERROR, SPAN)
def add_context_lines_to_frames(client, event):
    # divide frames up into source files before reading from disk. This should help
    # with utilizing the disk cache better
    #
    # TODO: further optimize by only opening each file once and reading all needed source
    # TODO: blocks at once.
    per_file = defaultdict(list)
    _process_stack_frames(
        event,
        lambda frame: per_file[frame["context_metadata"][0]].append(frame) if "context_metadata" in frame else None,
    )
    for frames in per_file.values():
        for frame in frames:
            # context_metadata key has been set in elasticapm.utils.stacks.get_frame_info for
            # all frames for which we should gather source code context lines
            fname, lineno, context_lines, loader, module_name = frame.pop("context_metadata")
            pre_context, context_line, post_context = get_lines_from_file(
                fname, lineno, context_lines, loader, module_name
            )
            if context_line:
                frame["pre_context"] = pre_context
                frame["context_line"] = context_line
                frame["post_context"] = post_context
    return event


@for_events(ERROR, SPAN)
def mark_in_app_frames(client, event):
    warnings.warn(
        "The mark_in_app_frames processor is deprecated and can be removed from your PROCESSORS setting",
        DeprecationWarning,
    )
    return event


def _sanitize(key, value, **kwargs):
    if "sanitize_field_names" in kwargs:
        sanitize_field_names = kwargs["sanitize_field_names"]
    else:
        sanitize_field_names = BASE_SANITIZE_FIELD_NAMES

    if value is None:
        return

    if isinstance(value, dict):
        # varmap will call _sanitize on each k:v pair of the dict, so we don't
        # have to do anything with dicts here
        return value

    if not key:  # key can be a NoneType
        return value

    key = key.lower()
    for field in sanitize_field_names:
        if field.match(key.strip()):
            # store mask as a fixed length for security
            return MASK
    return value


def _sanitize_string(unsanitized, itemsep, kvsep, sanitize_field_names=BASE_SANITIZE_FIELD_NAMES):
    """
    sanitizes a string that contains multiple key/value items
    :param unsanitized: the unsanitized string
    :param itemsep: string that separates items
    :param kvsep: string that separates key from value
    :param sanitize_field_names: field names to pass to _sanitize
    :return: a sanitized string
    """
    sanitized = []
    kvs = unsanitized.split(itemsep)
    for kv in kvs:
        kv = kv.split(kvsep)
        if len(kv) == 2:
            sanitized.append((kv[0], _sanitize(kv[0], kv[1], sanitize_field_names=sanitize_field_names)))
        else:
            sanitized.append(kv)
    return itemsep.join(kvsep.join(kv) for kv in sanitized)


def _process_stack_frames(event, func):
    if "stacktrace" in event:
        for frame in event["stacktrace"]:
            func(frame)
    # an error can have two stacktraces, one in "exception", one in "log"
    if "exception" in event and "stacktrace" in event["exception"]:
        for frame in event["exception"]["stacktrace"]:
            func(frame)
        # check for chained exceptions
        cause = event["exception"].get("cause", None)
        while cause:
            if "stacktrace" in cause[0]:
                for frame in cause[0]["stacktrace"]:
                    func(frame)
            cause = cause[0].get("cause", None)
    if "log" in event and "stacktrace" in event["log"]:
        for frame in event["log"]["stacktrace"]:
            func(frame)
    return event



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import os
import uuid

import pytest

from elasticapm.conf import constants

azureblob = pytest.importorskip("azure.storage.blob")
azurequeue = pytest.importorskip("azure.storage.queue")
azuretable = pytest.importorskip("azure.cosmosdb.table")
azuredatatable = pytest.importorskip("azure.data.tables")
azurefile = pytest.importorskip("azure.storage.fileshare")
pytestmark = [pytest.mark.azurestorage]


from azure.cosmosdb.table.tableservice import TableService
from azure.data.tables import TableServiceClient as DataTableServiceClient
from azure.storage.blob import BlobServiceClient
from azure.storage.fileshare import ShareClient
from azure.storage.queue import QueueClient

CONNECTION_STRING = os.getenv("AZURE_STORAGE_CONNECTION_STRING")

if not CONNECTION_STRING:
    pytestmark.append(
        pytest.mark.skip("Skipping azure storage tests, no AZURE_STORAGE_CONNECTION_STRING environment variable set")
    )


@pytest.fixture()
def container_client(blob_service_client):
    container_name = "apm-agent-python-ci-" + str(uuid.uuid4())
    container_client = blob_service_client.create_container(container_name)

    yield container_client

    blob_service_client.delete_container(container_name)


@pytest.fixture()
def blob_service_client():
    blob_service_client = BlobServiceClient.from_connection_string(CONNECTION_STRING)
    return blob_service_client


@pytest.fixture()
def queue_client():
    queue_name = "apm-agent-python-ci-" + str(uuid.uuid4())
    queue_client = QueueClient.from_connection_string(CONNECTION_STRING, queue_name)
    queue_client.create_queue()

    yield queue_client

    queue_client.delete_queue()


@pytest.fixture()
def data_table_service():
    table_name = "apmagentpythonci" + str(uuid.uuid4().hex)
    data_table_service_client = DataTableServiceClient.from_connection_string(conn_str=CONNECTION_STRING)
    data_table_service = data_table_service_client.get_table_client(table_name)
    data_table_service.create_table()
    data_table_service.table_name = table_name

    yield data_table_service

    data_table_service.delete_table()


@pytest.fixture()
def table_service():
    table_name = "apmagentpythonci" + str(uuid.uuid4().hex)
    table_service = TableService(connection_string=CONNECTION_STRING)
    table_service.create_table(table_name)
    table_service.table_name = table_name

    yield table_service

    table_service.delete_table(table_name)


@pytest.fixture()
def share_client():
    share_name = "apmagentpythonci" + str(uuid.uuid4().hex)
    share_client = ShareClient.from_connection_string(conn_str=CONNECTION_STRING, share_name=share_name)
    share_client.create_share()

    yield share_client

    share_client.delete_share()


def test_blob_list_blobs(instrument, elasticapm_client, container_client):
    elasticapm_client.begin_transaction("transaction.test")
    list(container_client.list_blobs())
    elasticapm_client.end_transaction("MyView")
    span = elasticapm_client.events[constants.SPAN][0]

    assert span["name"] == "AzureBlob ListBlobs {}".format(container_client.container_name)
    assert span["type"] == "storage"
    assert span["subtype"] == "azureblob"
    assert span["action"] == "ListBlobs"


def test_blob_create_container(instrument, elasticapm_client, blob_service_client):
    elasticapm_client.begin_transaction("transaction.test")
    container_name = str(uuid.uuid4())
    container_client = blob_service_client.create_container(container_name)
    blob_service_client.delete_container(container_name)
    elasticapm_client.end_transaction("MyView")
    span = elasticapm_client.events[constants.SPAN][0]

    assert span["name"] == "AzureBlob Create {}".format(container_name)
    assert span["type"] == "storage"
    assert span["subtype"] == "azureblob"
    assert span["action"] == "Create"


def test_blob_upload(instrument, elasticapm_client, container_client, blob_service_client):
    elasticapm_client.begin_transaction("transaction.test")
    # Upload this file to the container
    blob_client = blob_service_client.get_blob_client(container=container_client.container_name, blob=__file__)
    with open(__file__, "rb") as data:
        blob_client.upload_blob(data)
    elasticapm_client.end_transaction("MyView")
    span = elasticapm_client.events[constants.SPAN][0]

    assert span["name"] == "AzureBlob Upload {}/{}".format(container_client.container_name, __file__)
    assert span["type"] == "storage"
    assert span["subtype"] == "azureblob"
    assert span["action"] == "Upload"


def test_queue(instrument, elasticapm_client, queue_client):
    elasticapm_client.begin_transaction("transaction.test")
    # Send a message
    queue_client.send_message("Test message")
    list(queue_client.peek_messages())
    messages = queue_client.receive_messages()
    for msg_batch in messages.by_page():
        for msg in msg_batch:
            queue_client.delete_message(msg)
    elasticapm_client.end_transaction("MyView")

    span = elasticapm_client.events[constants.SPAN][0]
    assert span["name"] == "AzureQueue SEND to {}".format(queue_client.queue_name)
    assert span["type"] == "messaging"
    assert span["subtype"] == "azurequeue"
    assert span["action"] == "send"

    span = elasticapm_client.events[constants.SPAN][1]
    assert span["name"] == "AzureQueue PEEK from {}".format(queue_client.queue_name)
    assert span["type"] == "messaging"
    assert span["subtype"] == "azurequeue"
    assert span["action"] == "peek"

    span = elasticapm_client.events[constants.SPAN][2]
    assert span["name"] == "AzureQueue RECEIVE from {}".format(queue_client.queue_name)
    assert span["type"] == "messaging"
    assert span["subtype"] == "azurequeue"
    assert span["action"] == "receive"

    span = elasticapm_client.events[constants.SPAN][3]
    assert span["name"] == "AzureQueue DELETE from {}".format(queue_client.queue_name)
    assert span["type"] == "messaging"
    assert span["subtype"] == "azurequeue"
    assert span["action"] == "delete"


def test_data_table_create(instrument, elasticapm_client):
    table_name = "apmagentpythonci" + str(uuid.uuid4().hex)
    data_table_service_client = DataTableServiceClient.from_connection_string(conn_str=CONNECTION_STRING)
    data_table_service = data_table_service_client.get_table_client(table_name)

    elasticapm_client.begin_transaction("transaction.test")
    data_table_service.create_table()
    data_table_service.delete_table()
    elasticapm_client.end_transaction("MyView")

    span = elasticapm_client.events[constants.SPAN][0]

    assert span["name"] == "AzureTable Create {}".format(table_name)
    assert span["type"] == "storage"
    assert span["subtype"] == "azuretable"
    assert span["action"] == "Create"


def test_table_create(instrument, elasticapm_client):
    table_name = "apmagentpythonci" + str(uuid.uuid4().hex)
    table_service = TableService(connection_string=CONNECTION_STRING)

    elasticapm_client.begin_transaction("transaction.test")
    table_service.create_table(table_name)
    table_service.delete_table(table_name)
    elasticapm_client.end_transaction("MyView")

    span = elasticapm_client.events[constants.SPAN][0]

    assert span["name"] == "AzureTable Create {}".format(table_name)
    assert span["type"] == "storage"
    assert span["subtype"] == "azuretable"
    assert span["action"] == "Create"


def test_data_table(instrument, elasticapm_client, data_table_service):
    table_name = data_table_service.table_name
    elasticapm_client.begin_transaction("transaction.test")
    task = {"PartitionKey": "tasksSeattle", "RowKey": "001", "description": "Take out the trash", "priority": 200}
    data_table_service.create_entity(task)
    task = {"PartitionKey": "tasksSeattle", "RowKey": "001", "description": "Take out the garbage", "priority": 250}
    data_table_service.update_entity(task)
    task = data_table_service.get_entity("tasksSeattle", "001")
    data_table_service.delete_entity("tasksSeattle", "001")
    elasticapm_client.end_transaction("MyView")

    span = elasticapm_client.events[constants.SPAN][0]
    assert span["name"] == "AzureTable Insert {}".format(table_name)
    assert span["type"] == "storage"
    assert span["subtype"] == "azuretable"
    assert span["action"] == "Insert"

    span = elasticapm_client.events[constants.SPAN][1]
    assert span["name"] == "AzureTable Update {}(PartitionKey='tasksSeattle',RowKey='001')".format(table_name)
    assert span["type"] == "storage"
    assert span["subtype"] == "azuretable"
    assert span["action"] == "Update"

    span = elasticapm_client.events[constants.SPAN][2]
    assert span["name"] == "AzureTable Query {}(PartitionKey='tasksSeattle',RowKey='001')".format(table_name)
    assert span["type"] == "storage"
    assert span["subtype"] == "azuretable"
    assert span["action"] == "Query"

    span = elasticapm_client.events[constants.SPAN][3]
    assert span["name"] == "AzureTable Delete {}(PartitionKey='tasksSeattle',RowKey='001')".format(table_name)
    assert span["type"] == "storage"
    assert span["subtype"] == "azuretable"
    assert span["action"] == "Delete"


def test_table(instrument, elasticapm_client, table_service):
    table_name = table_service.table_name
    elasticapm_client.begin_transaction("transaction.test")
    task = {"PartitionKey": "tasksSeattle", "RowKey": "001", "description": "Take out the trash", "priority": 200}
    table_service.insert_entity(table_name, task)
    task = {"PartitionKey": "tasksSeattle", "RowKey": "001", "description": "Take out the garbage", "priority": 250}
    table_service.update_entity(table_name, task)
    task = table_service.get_entity(table_name, "tasksSeattle", "001")
    table_service.delete_entity(table_name, "tasksSeattle", "001")
    elasticapm_client.end_transaction("MyView")

    span = elasticapm_client.events[constants.SPAN][0]
    assert span["name"] == "AzureTable Insert {}".format(table_name)
    assert span["type"] == "storage"
    assert span["subtype"] == "azuretable"
    assert span["action"] == "Insert"

    span = elasticapm_client.events[constants.SPAN][1]
    assert span["name"] == "AzureTable Update {}(PartitionKey='tasksSeattle',RowKey='001')".format(table_name)
    assert span["type"] == "storage"
    assert span["subtype"] == "azuretable"
    assert span["action"] == "Update"

    span = elasticapm_client.events[constants.SPAN][2]
    assert span["name"] == "AzureTable Query {}(PartitionKey='tasksSeattle',RowKey='001')".format(table_name)
    assert span["type"] == "storage"
    assert span["subtype"] == "azuretable"
    assert span["action"] == "Query"

    span = elasticapm_client.events[constants.SPAN][3]
    assert span["name"] == "AzureTable Delete {}(PartitionKey='tasksSeattle',RowKey='001')".format(table_name)
    assert span["type"] == "storage"
    assert span["subtype"] == "azuretable"
    assert span["action"] == "Delete"


def test_fileshare(instrument, elasticapm_client, share_client):
    elasticapm_client.begin_transaction("transaction.test")
    # Upload this file to the share
    file_client = share_client.get_file_client("testfile.txt")
    with open(__file__, "rb") as data:
        file_client.upload_file(data)
    file_client.download_file()
    file_client.delete_file()
    elasticapm_client.end_transaction("MyView")

    span = elasticapm_client.events[constants.SPAN][0]
    assert span["name"] == "AzureFile Create {}/testfile.txt".format(share_client.share_name)
    assert span["type"] == "storage"
    assert span["subtype"] == "azurefile"
    assert span["action"] == "Create"

    span = elasticapm_client.events[constants.SPAN][1]
    assert span["name"] == "AzureFile Upload {}/testfile.txt".format(share_client.share_name)
    assert span["type"] == "storage"
    assert span["subtype"] == "azurefile"
    assert span["action"] == "Upload"

    span = elasticapm_client.events[constants.SPAN][2]
    assert span["name"] == "AzureFile Download {}/testfile.txt".format(share_client.share_name)
    assert span["type"] == "storage"
    assert span["subtype"] == "azurefile"
    assert span["action"] == "Download"

    span = elasticapm_client.events[constants.SPAN][3]
    assert span["name"] == "AzureFile Delete {}/testfile.txt".format(share_client.share_name)
    assert span["type"] == "storage"
    assert span["subtype"] == "azurefile"
    assert span["action"] == "Delete"



#  BSD 3-Clause License
#
#  Copyright (c) 2022, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


from elasticapm.instrumentation.packages.asyncio.base import AbstractInstrumentedModule
from elasticapm.utils.logging import get_logger

logger = get_logger("elasticapm.instrument")


class GRPCClientInstrumentation(AbstractInstrumentedModule):
    name = "grpc_client_instrumentation"
    creates_transactions = True
    instrument_list = [("grpc", "insecure_channel"), ("grpc", "secure_channel")]

    def call(self, module, method, wrapped, instance, args, kwargs):
        import grpc

        from elasticapm.contrib.grpc.client_interceptor import _ClientInterceptor

        result = wrapped(*args, **kwargs)
        target = kwargs.get("target") or args[0]
        if ":" in target:
            host, port = target.split(":")
            try:
                port = int(port)
            except ValueError:
                port = None
        else:
            host, port = target, None
        return grpc.intercept_channel(result, _ClientInterceptor(host, port, secure=method == "secure_channel"))


class GRPCServerInstrumentation(AbstractInstrumentedModule):
    name = "grpc_server_instrumentation"
    creates_transactions = True
    instrument_list = [("grpc", "server")]

    def call(self, module, method, wrapped, instance, args, kwargs):
        from elasticapm.contrib.grpc.server_interceptor import _ServerInterceptor

        interceptors = kwargs.get("interceptors") or (args[2] if len(args) > 2 else [])
        interceptors.insert(0, _ServerInterceptor())
        if len(args) > 2:
            args = list(args)
            args[2] = interceptors
        else:
            kwargs["interceptors"] = interceptors
        return wrapped(*args, **kwargs)


class GRPCAsyncServerInstrumentation(AbstractInstrumentedModule):
    name = "grpc_async_server_instrumentation"
    creates_transactions = True
    instrument_list = [("grpc.aio", "server")]

    def call(self, module, method, wrapped, instance, args, kwargs):
        from elasticapm.contrib.grpc.async_server_interceptor import _AsyncServerInterceptor

        interceptors = kwargs.get("interceptors") or (args[2] if len(args) > 2 else [])
        interceptors.insert(0, _AsyncServerInterceptor())
        if len(args) > 2:
            args = list(args)
            args[2] = interceptors
        else:
            kwargs["interceptors"] = interceptors
        return wrapped(*args, **kwargs)



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import absolute_import

import pytest
import werkzeug
from werkzeug.test import EnvironBuilder

from elasticapm.conf.constants import ERROR
from elasticapm.middleware import ElasticAPM


def example_app(environ, start_response):
    raise ValueError("hello world")


def test_error_handler(elasticapm_client):
    middleware = ElasticAPM(example_app, client=elasticapm_client)

    request = werkzeug.Request(environ=EnvironBuilder(path="an-error?foo=bar").get_environ())
    response = middleware(request.environ, lambda *args: None)

    with pytest.raises(ValueError):
        list(response)

    assert len(elasticapm_client.events) == 1
    event = elasticapm_client.events[ERROR][0]

    assert "exception" in event
    exc = event["exception"]
    assert exc["type"] == "ValueError"
    assert exc["message"] == "ValueError: hello world"

    assert "request" in event["context"]
    request = event["context"]["request"]
    assert request["url"]["full"] == "http://localhost/an-error?foo=bar"
    assert request["url"]["search"] == "?foo=bar"
    assert request["method"] == "GET"
    headers = request["headers"]
    assert "host" in headers, headers.keys()
    assert headers["host"] == "localhost"
    env = request["env"]
    assert "SERVER_NAME" in env, env.keys()
    assert env["SERVER_NAME"] == "localhost"
    assert "SERVER_PORT" in env, env.keys()
    assert env["SERVER_PORT"] == "80"



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import pytest  # isort:skip

from tests.instrumentation.elasticsearch_tests import get_kwargs

pytest.importorskip("elasticsearch._async")  # isort:skip

import json
import os
import urllib.parse

import pytest_asyncio
from elasticsearch import VERSION as ES_VERSION
from elasticsearch import AsyncElasticsearch

from elasticapm.conf.constants import TRANSACTION

pytestmark = [pytest.mark.elasticsearch, pytest.mark.asyncio, pytest.mark.integrationtest]

if "ES_URL" not in os.environ:
    pytestmark.append(pytest.mark.skip("Skipping elasticsearch test, no ES_URL environment variable"))


document_type = "_doc" if ES_VERSION[0] >= 6 else "doc"


@pytest_asyncio.fixture
async def async_elasticsearch(request):
    """AsyncElasticsearch client fixture."""
    client = AsyncElasticsearch(hosts=os.environ["ES_URL"])
    try:
        yield client
    finally:
        await client.indices.delete(index="*")
        await client.close()


async def test_ping(instrument, elasticapm_client, async_elasticsearch):
    elasticapm_client.begin_transaction("test")
    result = await async_elasticsearch.ping()
    elasticapm_client.end_transaction("test", "OK")
    parsed_url = urllib.parse.urlparse(os.environ["ES_URL"])

    transaction = elasticapm_client.events[TRANSACTION][0]
    spans = elasticapm_client.spans_for_transaction(transaction)
    assert len(spans) == 1
    span = spans[0]
    assert span["name"] == "ES HEAD /"
    assert span["type"] == "db"
    assert span["subtype"] == "elasticsearch"
    assert span["action"] == "query"
    assert span["sync"] is False
    assert span["context"]["destination"] == {
        "address": parsed_url.hostname,
        "port": parsed_url.port,
        "service": {"name": "", "resource": "elasticsearch", "type": ""},
    }
    assert span["context"]["http"]["status_code"] == 200


async def test_info(instrument, elasticapm_client, async_elasticsearch):
    elasticapm_client.begin_transaction("test")
    result = await async_elasticsearch.info()
    elasticapm_client.end_transaction("test", "OK")

    transaction = elasticapm_client.events[TRANSACTION][0]

    spans = elasticapm_client.spans_for_transaction(transaction)
    assert len(spans) == 1
    span = spans[0]
    assert span["name"] == "ES GET /"
    assert span["type"] == "db"
    assert span["subtype"] == "elasticsearch"
    assert span["action"] == "query"
    assert span["sync"] is False
    assert span["context"]["http"]["status_code"] == 200


async def test_create(instrument, elasticapm_client, async_elasticsearch):
    elasticapm_client.begin_transaction("test")
    responses = []
    iid = lambda: str(len(responses) + 1)
    if ES_VERSION[0] < 5:
        responses.append(
            await async_elasticsearch.create("tweets", document_type, {"user": "kimchy", "text": "hola"}, iid())
        )
    elif ES_VERSION[0] < 7:
        responses.append(
            await async_elasticsearch.create("tweets", document_type, iid(), body={"user": "kimchy", "text": "hola"})
        )
    elif ES_VERSION[0] < 8:
        responses.append(await async_elasticsearch.create("tweets", iid(), body={"user": "kimchy", "text": "hola"}))
    else:
        pass  # elasticsearch-py 8+ doesn't support positional arguments
    responses.append(
        await async_elasticsearch.create(index="tweets", id=iid(), **get_kwargs({"user": "kimchy", "text": "hola"}))
    )
    elasticapm_client.end_transaction("test", "OK")

    transaction = elasticapm_client.events[TRANSACTION][0]

    spans = elasticapm_client.spans_for_transaction(transaction)
    assert len(spans) == len(responses)

    for i, span in enumerate(spans):
        if ES_VERSION[0] >= 5:
            assert span["name"] in (
                "ES PUT /tweets/%s/%d/_create" % (document_type, i + 1),
                "ES PUT /tweets/_create/%d" % (i + 1),
            )
        else:
            assert span["name"] == "ES PUT /tweets/%s/%d" % (document_type, i + 1)
        assert span["type"] == "db"
        assert span["subtype"] == "elasticsearch"
        assert span["action"] == "query"
        assert span["context"]["db"]["type"] == "elasticsearch"
        assert "statement" not in span["context"]["db"]
        assert span["context"]["http"]["status_code"] == 201


async def test_search_body(instrument, elasticapm_client, async_elasticsearch):
    await async_elasticsearch.create(
        index="tweets", id="1", refresh=True, **get_kwargs({"user": "kimchy", "text": "hola", "userid": 1})
    )
    elasticapm_client.begin_transaction("test")
    search_query = {"query": {"term": {"user": "kimchy"}}, "sort": ["userid"]}
    result = await async_elasticsearch.search(body=search_query, params=None)
    elasticapm_client.end_transaction("test", "OK")

    transaction = elasticapm_client.events[TRANSACTION][0]
    assert result["hits"]["hits"][0]["_source"] == {"user": "kimchy", "text": "hola", "userid": 1}
    spans = elasticapm_client.spans_for_transaction(transaction)
    assert len(spans) == 1
    span = spans[0]
    # Depending on ES_VERSION, could be /_all/_search or /_search, and GET or POST
    assert span["name"] in ("ES GET /_search", "ES GET /_all/_search", "ES POST /_search")
    assert span["type"] == "db"
    assert span["subtype"] == "elasticsearch"
    assert span["action"] == "query"
    assert span["context"]["db"]["type"] == "elasticsearch"
    assert json.loads(span["context"]["db"]["statement"]) == json.loads(
        '{"sort": ["userid"], "query": {"term": {"user": "kimchy"}}}'
    ) or json.loads(span["context"]["db"]["statement"]) == json.loads(
        '{"query": {"term": {"user": "kimchy"}}, "sort": ["userid"]}'
    )
    assert span["sync"] is False
    if ES_VERSION[0] >= 7:
        assert span["context"]["db"]["rows_affected"] == 1
    assert span["context"]["http"]["status_code"] == 200


async def test_count_body(instrument, elasticapm_client, async_elasticsearch):
    await async_elasticsearch.create(
        index="tweets", id="1", refresh=True, **get_kwargs({"user": "kimchy", "text": "hola"})
    )
    elasticapm_client.begin_transaction("test")
    search_query = {"query": {"term": {"user": "kimchy"}}}
    result = await async_elasticsearch.count(body=search_query)
    elasticapm_client.end_transaction("test", "OK")

    transaction = elasticapm_client.events[TRANSACTION][0]
    assert result["count"] == 1
    spans = elasticapm_client.spans_for_transaction(transaction)
    assert len(spans) == 1
    span = spans[0]
    assert span["name"] in ("ES GET /_count", "ES POST /_count", "ES GET /_all/_count")
    assert span["type"] == "db"
    assert span["subtype"] == "elasticsearch"
    assert span["action"] == "query"
    assert span["context"]["db"]["type"] == "elasticsearch"
    assert json.loads(span["context"]["db"]["statement"]) == json.loads('{"query": {"term": {"user": "kimchy"}}}')
    assert span["sync"] is False
    assert span["context"]["http"]["status_code"] == 200



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: test.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\ntest.proto\x12\x04test\"\x1a\n\x07Message\x12\x0f\n\x07message\x18\x01 \x01(\t\"4\n\x0fMessageResponse\x12\x0f\n\x07message\x18\x01 \x01(\t\x12\x10\n\x08received\x18\x02 \x01(\x08\x32\x9a\x02\n\x0bTestService\x12;\n\x11GetServerResponse\x12\r.test.Message\x1a\x15.test.MessageResponse\"\x00\x12@\n\x16GetServerResponseAbort\x12\r.test.Message\x1a\x15.test.MessageResponse\"\x00\x12\x46\n\x1cGetServerResponseUnavailable\x12\r.test.Message\x1a\x15.test.MessageResponse\"\x00\x12\x44\n\x1aGetServerResponseException\x12\r.test.Message\x1a\x15.test.MessageResponse\"\x00\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'test_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _MESSAGE._serialized_start=20
  _MESSAGE._serialized_end=46
  _MESSAGERESPONSE._serialized_start=48
  _MESSAGERESPONSE._serialized_end=100
  _TESTSERVICE._serialized_start=103
  _TESTSERVICE._serialized_end=385
# @@protoc_insertion_point(module_scope)



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import pytest  # isort:skip

eventlet = pytest.importorskip("eventlet")  # isort:skip

from eventlet.patcher import is_monkey_patched

import elasticapm.context
from elasticapm.context.threadlocal import ThreadLocalContext

pytestmark = pytest.mark.eventlet


def test_eventlet_thread_monkeypatched():
    eventlet.monkey_patch(thread=True)
    assert is_monkey_patched("thread")

    execution_context = elasticapm.context.init_execution_context()

    # Should always use ThreadLocalContext when eventlet has patched
    # threading.local
    assert isinstance(execution_context, ThreadLocalContext)



#  BSD 3-Clause License
#
#  Copyright (c) 2021, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import pytest

import elasticapm
from elasticapm.conf.constants import SPAN, TRANSACTION


@pytest.mark.parametrize(
    "elasticapm_client",
    [
        {
            "span_compression_enabled": True,
            "span_compression_same_kind_max_duration": "5ms",
            "span_compression_exact_match_max_duration": "5ms",
        }
    ],
    indirect=True,
)
def test_exact_match(elasticapm_client):
    transaction = elasticapm_client.begin_transaction("test")
    with elasticapm.capture_span(
        "test",
        span_type="a",
        span_subtype="b",
        span_action="c",
        leaf=True,
        duration=0.002,
        extra={"destination": {"service": {"resource": "x"}}},
    ) as span1:
        assert span1.is_compression_eligible()
    with elasticapm.capture_span(
        "test",
        span_type="a",
        span_subtype="b",
        span_action="c",
        leaf=True,
        duration=0.003,
        extra={"destination": {"service": {"resource": "x"}}},
    ) as span2:
        pass
    assert span2.is_compression_eligible()
    assert span1.is_exact_match(span2)
    elasticapm_client.end_transaction("test")
    spans = elasticapm_client.events[SPAN]
    assert len(spans) == 1
    span = spans[0]
    assert "composite" in span
    assert span["composite"]["count"] == 2
    assert span["composite"]["sum"] == 5
    assert span["composite"]["compression_strategy"] == "exact_match"


@pytest.mark.parametrize(
    "elasticapm_client",
    [
        {
            "span_compression_enabled": True,
            "span_compression_same_kind_max_duration": "5ms",
            "span_compression_exact_match_max_duration": "5ms",
        }
    ],
    indirect=True,
)
def test_same_kind(elasticapm_client):
    transaction = elasticapm_client.begin_transaction("test")
    with elasticapm.capture_span(
        "test1",
        span_type="a",
        span_subtype="b",
        span_action="c",
        leaf=True,
        duration=0.002,
        extra={"destination": {"service": {"resource": "x"}}},
    ) as span1:
        assert span1.is_compression_eligible()
    with elasticapm.capture_span(
        "test2",
        span_type="a",
        span_subtype="b",
        span_action="c",
        leaf=True,
        duration=0.003,
        extra={"destination": {"service": {"resource": "x"}}},
    ) as span2:
        pass
    assert span2.is_compression_eligible()
    assert not span1.is_exact_match(span2)
    assert span1.is_same_kind(span2)
    elasticapm_client.end_transaction("test")
    spans = elasticapm_client.events[SPAN]
    assert len(spans) == 1
    span = spans[0]

    assert span["name"] == "Calls to x"
    assert "composite" in span
    assert span["composite"]["count"] == 2
    assert span["composite"]["sum"] == 5
    assert span["composite"]["compression_strategy"] == "same_kind"


@pytest.mark.parametrize(
    "elasticapm_client",
    [
        {
            "span_compression_enabled": True,
            "span_compression_same_kind_max_duration": "5ms",
            "span_compression_exact_match_max_duration": "5ms",
        }
    ],
    indirect=True,
)
def test_exact_match_after_same_kind(elasticapm_client):
    # if a span that is an exact match is attempted to be compressed with a same_kind composite, it stays same_kind
    transaction = elasticapm_client.begin_transaction("test")
    with elasticapm.capture_span(
        "test1",
        span_type="a",
        span_subtype="b",
        span_action="c",
        leaf=True,
        duration=0.002,
        extra={"destination": {"service": {"resource": "x"}}},
    ) as span1:
        assert span1.is_compression_eligible()
    with elasticapm.capture_span(
        "test2",
        span_type="a",
        span_subtype="b",
        span_action="c",
        leaf=True,
        duration=0.003,
        extra={"destination": {"service": {"resource": "x"}}},
    ) as span2:
        pass
    assert span2.is_compression_eligible()
    assert not span1.is_exact_match(span2)
    assert span1.is_same_kind(span2)
    with elasticapm.capture_span(
        "test1",
        span_type="a",
        span_subtype="b",
        span_action="c",
        leaf=True,
        duration=0.002,
        extra={"destination": {"service": {"resource": "x"}}},
    ) as span3:
        assert span3.is_compression_eligible()
    elasticapm_client.end_transaction("test")
    spans = elasticapm_client.events[SPAN]
    assert len(spans) == 1
    span = spans[0]
    assert span["composite"]["compression_strategy"] == "same_kind"
    assert span["composite"]["count"] == 3


@pytest.mark.parametrize(
    "elasticapm_client",
    [
        {
            "span_compression_enabled": True,
            "span_compression_same_kind_max_duration": "5ms",
            "span_compression_exact_match_max_duration": "5ms",
        }
    ],
    indirect=True,
)
def test_nested_spans(elasticapm_client):
    transaction = elasticapm_client.begin_transaction("test")
    with elasticapm.capture_span("test", "x.y.z") as span1:
        with elasticapm.capture_span(
            "test1",
            span_type="a",
            span_subtype="b",
            span_action="c",
            leaf=True,
            duration=0.002,
            extra={"destination": {"service": {"resource": "x"}}},
        ) as span2:
            pass
        with elasticapm.capture_span(
            "test2",
            span_type="a",
            span_subtype="b",
            span_action="c",
            leaf=True,
            duration=0.002,
            extra={"destination": {"service": {"resource": "x"}}},
        ) as span3:
            pass
        assert span1.compression_buffer is span2
        assert span2.composite
    # assert transaction.compression_buffer is span1
    # assert not span1.compression_buffer
    elasticapm_client.end_transaction("test")
    spans = elasticapm_client.events[SPAN]
    assert len(spans) == 2


@pytest.mark.parametrize(
    "elasticapm_client",
    [
        {
            "span_compression_enabled": True,
            "span_compression_same_kind_max_duration": "5ms",
            "span_compression_exact_match_max_duration": "5ms",
        }
    ],
    indirect=True,
)
def test_buffer_is_reported_if_next_child_ineligible(elasticapm_client):
    transaction = elasticapm_client.begin_transaction("test")
    with elasticapm.capture_span("test", "x.y.z") as span1:
        with elasticapm.capture_span(
            "test",
            "x.y.z",
            leaf=True,
            duration=0.002,
            extra={"destination": {"service": {"resource": "x"}}},
        ) as span2:
            pass
        assert span1.compression_buffer is span2
        with elasticapm.capture_span("test", "x.y.z") as span3:
            pass
        assert span1.compression_buffer is None
    elasticapm_client.end_transaction("test")
    spans = elasticapm_client.events[SPAN]
    assert len(spans) == 3


@pytest.mark.parametrize(
    "elasticapm_client",
    [
        {
            "span_compression_enabled": True,
            "span_compression_same_kind_max_duration": "5ms",
            "span_compression_exact_match_max_duration": "5ms",
        }
    ],
    indirect=True,
)
def test_compressed_spans_not_counted(elasticapm_client):
    t = elasticapm_client.begin_transaction("test")
    assert t.config_span_compression_enabled
    assert t.config_span_compression_exact_match_max_duration.total_seconds() == 0.005
    assert t.config_span_compression_same_kind_max_duration.total_seconds() == 0.005
    with elasticapm.capture_span(
        "test1",
        span_type="a",
        span_subtype="b",
        span_action="c",
        leaf=True,
        duration=0.002,
        extra={"destination": {"service": {"resource": "x"}}},
    ) as span1:
        pass
    with elasticapm.capture_span(
        "test2",
        span_type="a",
        span_subtype="b",
        span_action="c",
        leaf=True,
        duration=0.003,
        extra={"destination": {"service": {"resource": "x"}}},
    ) as span2:
        pass
    elasticapm_client.end_transaction("test")
    assert len(elasticapm_client.events[TRANSACTION]) == 1
    transaction = elasticapm_client.events[TRANSACTION][0]
    spans = elasticapm_client.events[SPAN]
    assert len(spans) == 1
    assert transaction["span_count"]["started"] == 1
    assert transaction["span_count"]["dropped"] == 0


@pytest.mark.parametrize(
    "elasticapm_client",
    [
        {
            "span_compression_enabled": False,
            "span_compression_same_kind_max_duration": "5ms",
            "span_compression_exact_match_max_duration": "5ms",
        }
    ],
    indirect=True,
)
def test_span_compression_disabled(elasticapm_client):
    transaction = elasticapm_client.begin_transaction("test")
    with elasticapm.capture_span(
        "test",
        span_type="a",
        span_subtype="b",
        span_action="c",
        leaf=True,
        duration=2,
        extra={"destination": {"service": {"resource": "x"}}},
    ) as span1:
        assert not span1.is_compression_eligible()
    with elasticapm.capture_span(
        "test",
        span_type="a",
        span_subtype="b",
        span_action="c",
        leaf=True,
        duration=3,
        extra={"destination": {"service": {"resource": "x"}}},
    ) as span2:
        assert not span2.is_compression_eligible()
    elasticapm_client.end_transaction("test")
    spans = elasticapm_client.events[SPAN]
    assert len(spans) == 2
    span = spans[0]
    assert "composite" not in span



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import mock
import pytest

from elasticapm.utils import compat

try:
    from werkzeug.datastructures import MultiDict

    has_multidict = True
except ImportError:
    has_multidict = False

try:
    from django.utils.datastructures import MultiValueDict

    has_multivaluedict = True
except ImportError:
    has_multivaluedict = False


@mock.patch("platform.system")
@mock.patch("platform.python_implementation")
@mock.patch("platform.python_version_tuple")
def test_default_library_paths(version_tuple, python_implementation, system):
    cases = (
        ("Linux", ("3", "5", "1"), "CPython", ["*/lib/python3.5/*", "*/lib64/python3.5/*"]),
        ("Linux", ("2", "7", "9"), "CPython", ["*/lib/python2.7/*", "*/lib64/python2.7/*"]),
        ("Windows", ("3", "5", "1"), "CPython", ["*\\lib\\*"]),
        ("Windows", ("2", "7", "9"), "CPython", ["*\\lib\\*"]),
        ("Linux", ("3", "6", "3"), "PyPy", ["*/lib-python/3/*", "*/site-packages/*"]),
        ("Linux", ("2", "7", "9"), "PyPy", ["*/lib-python/2.7/*", "*/site-packages/*"]),
    )
    for system_name, version, implementation, expected in cases:
        system.return_value = system_name
        version_tuple.return_value = version
        python_implementation.return_value = implementation

        assert compat.get_default_library_patters() == expected


@pytest.mark.django
@pytest.mark.skipif(not has_multivaluedict, reason="Django not installed")
def test_multivalue_dict():
    d = MultiValueDict()
    d.update({"a": "b", "b": "d"})
    d.update({"a": "c", "e": "f"})
    d = compat.multidict_to_dict(d)
    assert d == {"a": ["b", "c"], "b": "d", "e": "f"}


@pytest.mark.flask
@pytest.mark.skipif(not has_multidict, reason="Werkzeug not installed")
def test_multi_dict():
    d = MultiDict()
    d.update({"a": "b", "b": "d"})
    d.update({"a": "c", "e": "f"})
    d = compat.multidict_to_dict(d)
    assert d == {"a": ["b", "c"], "b": "d", "e": "f"}



#!/usr/bin/env python

#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import importlib
import os
import sys
from os.path import abspath, dirname

try:
    import eventlet

    eventlet.monkey_patch()
except ImportError:
    pass

try:
    from psycopg2cffi import compat

    compat.register()
except ImportError:
    pass

where_am_i = dirname(abspath(__file__))


sys.path.insert(0, where_am_i)

# don't run tests of dependencies that land in "build" and "src"
collect_ignore = ["build", "src"]

pytest_plugins = ["tests.fixtures"]

for module, fixtures in {
    "django": "tests.contrib.django.fixtures",
    "flask": "tests.contrib.flask.fixtures",
    "aiohttp": "aiohttp.pytest_plugin",
    "sanic": "tests.contrib.sanic.fixtures",
}.items():
    try:
        importlib.import_module(module)
        pytest_plugins.append(fixtures)
    except ImportError:
        pass


def pytest_report_header(config):
    if "PYTHON_VERSION" in os.environ:
        return "matrix: {}/{}".format(os.environ.get("PYTHON_VERSION"), os.environ.get("FRAMEWORK"))
